@@ -1,5 +1,5 @@
 package Crypt::Argon2;
-$Crypt::Argon2::VERSION = '0.012';
+$Crypt::Argon2::VERSION = '0.014';
 use strict;
 use warnings;
 
@@ -7,34 +7,43 @@ use Exporter 5.57 'import';
 our @EXPORT_OK = qw/
 	argon2id_raw argon2id_pass argon2id_verify
 	argon2i_raw argon2i_pass argon2i_verify
-	argon2d_raw argon2_needs_rehash/;
+	argon2d_raw argon2_needs_rehash
+	argon2_verify argon2_crypt/;
 use XSLoader;
 XSLoader::load(__PACKAGE__, __PACKAGE__->VERSION || 0);
 
-use MIME::Base64 'decode_base64';
-
 my %multiplier = (
 	k => 1,
 	M => 1024,
 	G => 1024 * 1024,
 );
 
+my $regex = qr/ ^ \$ (argon2(?:i|d|id)) \$ v=(\d+) \$ m=(\d+), t=(\d+), p=(\d+) \$ ([^\$]+) \$ (.*) $ /x;
+
 sub argon2_needs_rehash {
 	my ($encoded, $type, $t_cost, $m_cost, $parallelism, $output_length, $salt_length) = @_;
 	$m_cost =~ s/ \A (\d+) ([kMG]) \z / $1 * $multiplier{$2} * 1024 /xmse;
 	$m_cost /= 1024;
-	my (undef, $name, $version, $argstring, $salt, $hash) = split /\$/, $encoded;
-	return 1 if $name ne $type;
-	return 1 if $version !~ /v=(\d+)/ or $1 != 19;
-	my %args;
-	while ($argstring =~ m/(\w)=(\d+)/gc) {
-		$args{$1} = $2;
-	}
-	return 1 if $args{t} < $t_cost or $args{m} < $m_cost or $args{p} < $parallelism;
-	return 1 if length decode_base64($salt) != $salt_length or length decode_base64($hash) != $output_length;
+	my ($name, $version, $m_got, $t_got, $parallel_got, $salt, $hash) = $encoded =~ $regex or return 1;
+	return 1 if $name ne $type or $version != 19 or $t_got != $t_cost or $m_got != $m_cost or $parallel_got != $parallelism;
+	return 1 if int(3 / 4 * length $salt) != $salt_length or int(3 / 4 * length $hash) != $output_length;
 	return 0;
 }
 
+sub argon2_verify {
+	my ($name) = $_[0] =~ $regex or return !!0;
+	my $verify = do { no strict; \&{"$name\_verify"} };
+	goto &{$verify};
+}
+
+sub argon2_crypt {
+	my ($password, $settings) = @_;
+	my ($name, $version, $m_got, $t_got, $parallel_got, $salt, $hash) = $settings =~ $regex or return undef;
+	my $length = length $hash ? int(3 / 4 * length $hash) : 16;
+	my $pass = do { no strict; \&{"$name\_pass"} };
+	return eval { $pass->($password, $salt, $t_got, $m_got, $parallel_got, $length) };
+}
+
 1;
 
 # ABSTRACT: Perl interface to the Argon2 key derivation functions
@@ -51,7 +60,7 @@ Crypt::Argon2 - Perl interface to the Argon2 key derivation functions
 
 =head1 VERSION
 
-version 0.012
+version 0.014
 
 =head1 SYNOPSIS
 
@@ -144,7 +153,15 @@ This function processes the C<$password> with the given C<$salt> and parameters
 
 =head2 argon2_needs_rehash($encoded, $type, $t_cost, $m_cost, $parallelism, $salt_length, $output_length)
 
-This function checks if a password-encoded string needs a rehash. It will return true if the C<$type> (valid values are C<argon2i>, C<argon2id> or C<argon2d>) mismatches or any of the C<$t_cost>, C<$m_cost>, C<$parallelism>, C<$salt_length> or C<$output_length> arguments are higher than in the password-encoded hash.
+This function checks if a password-encoded string needs a rehash. It will return true if the C<$type> (valid values are C<argon2i>, C<argon2id> or C<argon2d>), C<$t_cost>, C<$m_cost>, C<$parallelism>, C<$salt_length> or C<$output_length> arguments mismatches or any of the parameters of the password-encoded hash.
+
+=head2 argon2_verify($encoded, $password)
+
+This will verify the hash using C<argon2id_verify>, C<argon2i_verify> or C<argon2d_verify>, depending on the identifier in C<$encoded>.
+
+=head2 argon2_crypt($password, $settings)
+
+This function implements a C<crypt()> like interface to argon2. C<$password> is a password, but C<$settings> is a settings string (a password hash that may lack anything beyond the final C<$>).
 
 =head2 ACKNOWLEDGEMENTS
 
